/*

This example finds unique VAC driver interface, appends a Native Mode key
and opens the driver in the Native Mode.

Can be built using MS VC++ 2005 (compiler version 14.x) and higher.

Don't forget to replace GUIDs in params.h.

To build an executable:

cl vacnm.cpp

You can use DDK build utility instead, or create an IDE project.

*/



#define _CRT_SECURE_NO_WARNINGS
#define _CRT_SECURE_NO_DEPRECATE

#include <windows.h>
#include <setupapi.h>

#include "params.h"

#define _StrX(x) #x
#define _S(x) _StrX (x)





// ###################################################################
//
// Default library definition
// ==========================
//
// ###################################################################





#pragma comment (lib, "kernel32.lib")
#pragma comment (lib, "user32.lib")
#pragma comment (lib, "setupapi.lib")





// ###################################################################
//
// Variables
// =========
//
// ###################################################################





UINT DebugPrintLevel = 5;





//
// Debug print with arglist
// ========================
//

void VDebug (UINT Level, PCSTR Fmt, va_list List) {

  char Str [MAX_PATH];

  strcpy (Str, (Level >= 9)? "*" : "");

  wvsprintf (Str, Fmt, List);

  strcat (Str, "\n");

  OutputDebugString (Str);

}





//
// Debug print with variable arguments
// ===================================
//

void _cdecl Debug (UINT Level, PCSTR Fmt, ...) {

  va_list List;

  if (Level >= DebugPrintLevel) {

    va_start (List, Fmt);
    VDebug (Level, Fmt, List);
    va_end (List);

  }

}





static void Message (PCSTR Format, va_list Args, UINT Type) {

  char Str [256];

  wvsprintf (Str, Format, Args);

  Debug ((Type & MB_ICONERROR)? 9 : 8, "Msg: %.200s", Str);

  MessageBox (HWND_DESKTOP, Str, "VAC Native Mode example", Type);
  
}





static inline void _cdecl Error (PCSTR Format, ...) {

  va_list Args;

  va_start (Args, Format);
  Message (Format, Args, MB_ICONERROR | MB_OK);
  va_end (Args);
  
}





static inline void _cdecl Info (PCSTR Format, ...) {

  va_list Args;

  va_start (Args, Format);
  Message (Format, Args, MB_ICONINFORMATION | MB_OK);
  va_end (Args);
  
}





void TranslateSetupApiError (void) {

  DWORD const OldCode = GetLastError ();

  DWORD const Mask = APPLICATION_ERROR_MASK | ERROR_SEVERITY_ERROR;

  if ((OldCode & Mask) == Mask) {

    DWORD const NewCode = HRESULT_FROM_SETUPAPI (OldCode);

    if (NewCode != OldCode) {

      SetLastError (NewCode);

    }

  }

}





static PCSTR ErrorText (DWORD Code) {

  static char Text [256];

  DWORD const Err = GetLastError ();

  wsprintf (Text, "%u/%u - ", HRESULT_FACILITY (Code), HRESULT_CODE (Code));

  UINT Offset = strlen (Text);

  char *cp = Text + Offset;
  UINT Avail = sizeof (Text) - Offset - 1;

  FormatMessage (
    FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_MAX_WIDTH_MASK,
    NULL,
    Code,
    MAKELANGID (LANG_NEUTRAL, SUBLANG_DEFAULT),
    cp,
    Avail,
    NULL 
  );

  cp += strlen (cp) - 1;

  if (*cp == ' ') {

    *cp = '\0';

  }

  SetLastError (Err);

  return Text;

}





static inline PCSTR LastErrorText (void) {

  return ErrorText (GetLastError ());

}




static PCSTR GuidToStr (PSTR Buffer, GUID const &Guid) {

  wsprintf (
    Buffer,
    "%08x-%04x-%04x-%02x%02x-%02x%02x%02x%02x%02x%02x",
    Guid.Data1,
    Guid.Data2, Guid.Data3,
    Guid.Data4 [0], Guid.Data4 [1],
    Guid.Data4 [2], Guid.Data4 [3], Guid.Data4 [4],
    Guid.Data4 [5], Guid.Data4 [6], Guid.Data4 [7]
  );

  return Buffer;

}





static HANDLE OpenVac4InNativeMode (void) {

  HANDLE Dev = 0;

  HDEVINFO DevList = 0;

  do {

    // Define a compiler-friendly GUID attribyte (driver interface)

    GUID const ClassGuid = __uuidof (ProductGuid);

    // Get a list of devices registered in the class

    char Class [40];

    Debug (7, "Querying devices in class %s", GuidToStr (Class, ClassGuid));

    DevList = SetupDiGetClassDevs (
      &ClassGuid,
      0,
      0,
      DIGCF_DEVICEINTERFACE | DIGCF_PRESENT
    );

    if (DevList == INVALID_HANDLE_VALUE) {

      TranslateSetupApiError ();

      Error ("Cannot get class device list (%s)", LastErrorText ());

      DevList = 0;

      break;

    }

    SP_DEVICE_INTERFACE_DATA IfData;

    ZeroMemory (&IfData, sizeof (IfData));

    IfData.cbSize = sizeof (IfData);

    BOOL Res;

    // Get first interface data

    Debug (7, "Querying first interface");

    Res = SetupDiEnumDeviceInterfaces (DevList, 0, &ClassGuid, 0, &IfData);

    if (!Res) {

      TranslateSetupApiError ();

      if (GetLastError () == ERROR_NO_MORE_ITEMS) {

        Error ("Driver not found");

      } else {

        Error ("Cannot get interface data (%s)", LastErrorText ());

      }

      break;

    }

    struct IfDetail : public SP_DEVICE_INTERFACE_DETAIL_DATA {

      BYTE VarPart [MAX_PATH];
    
    } IfDetData;

    ZeroMemory (&IfDetData, sizeof (IfDetData));

    IfDetData.cbSize = sizeof (SP_DEVICE_INTERFACE_DETAIL_DATA);

    // Get interface details

    Res = SetupDiGetDeviceInterfaceDetail (
      DevList,
      &IfData,
      &IfDetData,
      sizeof (IfDetData),
      0,
      0
    );

    if (!Res) {

      TranslateSetupApiError ();

      Error ("Cannot get interface details (%s)", LastErrorText ());

      break;

    }

    Debug (7, "Found driver: %s", IfDetData.DevicePath);

    GUID const Key = __uuidof (NativeModeKeyGuid);

    // Append the Native Mode key to an interface name

    strcat (IfDetData.DevicePath, "\\");

    GuidToStr (IfDetData.DevicePath + strlen (IfDetData.DevicePath), Key);

    // Open VAC driver in the Native Mode

    Debug (7, "Opening %s", IfDetData.DevicePath);

    Dev = CreateFile (
      IfDetData.DevicePath, GENERIC_READ, 0, 0, OPEN_EXISTING, 0, 0
    );

    if (Dev == INVALID_HANDLE_VALUE) {

      Dev = 0;

      if (GetLastError () == ERROR_FILE_NOT_FOUND) {

        Error ("Driver does not accept the key");

      } else {

        Error ("Cannot open %s (%s)", IfDetData.DevicePath, LastErrorText ());

      }

      break;

    }

#pragma warning (push)
#pragma warning (disable: 4127 /* Conditional expression is constant */)

  // Form a light-weight try-finally block

  } while (false);

#pragma warning (pop)

  if (DevList) {

    SetupDiDestroyDeviceInfoList (DevList);

  }

  return Dev;

}





static int Main (void) {

  HANDLE const H = OpenVac4InNativeMode ();

  if (H) {

    Info ("VAC driver is opened in the Native Mode");

    CloseHandle (H);

  }

  return !H;

}





int WINAPI WinMain (
  HINSTANCE Instance,
  HINSTANCE PrevInstance,
  LPSTR CmdLine,
  int CmdShow
) {

  // Make arguments "referenced"

  (void) (Instance, PrevInstance, CmdLine, CmdShow);

  return Main ();

}
